require 'lusi_api/core/code'
require 'lusi_api/core/util'
require 'lusi_api/core/xml'


module LUSI
  module API
    module Calendar


      # Represents an academic week in the LUSI API
      class Week < LUSI::API::Core::Code

        # @!attribute [rw] att_identity
        #   @return [Integer, nil] the academic timetable (ATT) week identity
        attr_accessor :att_identity

        # @!attribute [rw] monday_date
        #   @return [DateTime, nil] the Monday starting the week
        attr_accessor :monday_date

        # @!attribute [rw] term
        #   @return [String, nil] The term/period the week is in
        attr_accessor :term

        # @!attribute [rw] year
        #   @return [LUSI::API::Calendar::Year, nil] the academic year containing the week
        attr_accessor :year

        # Returns a Week instance for the current academic week
        # @param api [LUSI::API::Core::API] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @return [LUSI::API::Calendar::Week] the current academic week
        # @yield [obj] Passes the Week instance to the block
        # @yieldparam obj [LUSI::API::Calendar::Week] the Week instance
        def self.get_current_academic_week(api = nil, lookup = nil)
          xml = api.call('LUSIReference', 'General.asmx', 'GetCurrentAcademicWeek')
          obj = new(LUSI::API::Core::XML.xml_at(xml, 'xmlns:Week'), lookup)
          yield(obj) if block_given?
          obj
        end

        # Returns Week instances for all or one specific academic year
        # @param api [LUSI::API::Core::API] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param use_lookup [Boolean] if true, use the lookup service to identify instances before trying the API
        # @param year [LUSI::API::Calendar::Year, String, nil] the academic year required (all defined weeks if nil)
        # @return [Array<LUSI::API::Calendar::Week>] the selected acadmic weeks
        # @yield [obj] Passes the Week instance to the block
        # @yieldparam obj [LUSI::API::Calendar::Week] the Week instance
        def self.get_instance(api = nil, lookup = nil, use_lookup: true, year_identity: nil)
          # Search the lookup service
          if lookup and use_lookup
            week = lookup.lookup(:week, year_identity)
            if week
              yield(week) if block_given?
              return week
            end
          end
          # Search the API
          super(api, lookup, 'LUSIReference', 'General.asmx', 'GetWeeks', 'xmlns:Week', year_identity: year_identity)
        end

        # Initialises a new Week instance
        # @param (see LUSI::API::Core::Code#initialize)
        # @param att_identity [Integer, nil] the default academic timetable (ATT) identity
        # @param monday_date [DateTime, nil] the default Monday week-start date
        # @param term [String, nil] the default term
        # @param year [LUSI::API::Calendar::Year, nil] the default academic year
        # @return [void]
        def initialize(xml = nil, lookup = nil, term: nil, monday_date: nil, year: nil, att_identity: nil, **kwargs)
          super(xml, lookup, **kwargs)
          @att_identity = LUSI::API::Core::XML.xml_int_at(xml, 'xmlns:ATTIdentity', att_identity)
          @monday_date = LUSI::API::Core::XML.xml_datetime_at(xml, 'xmlns:MondayDate', monday_date)
          @term = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Term', term)
          @year = LUSI::API::Core::XML.lookup(xml, lookup, :year, 'xmlns:Year/xmlns:Identity', year)
        end

        protected

        def self.get_instance_params(**kwargs)
          year = kwargs[:year_identity] || ''
          year = year.identity if year.is_a?(Year)
          params = super(**kwargs)
          params[:YearIdentity] = year || ''
          params
        end

      end


      # Represents an academic year in the LUSI API
      class Year < LUSI::API::Core::Code

        extend LUSI::API::Core::Util

        # @!attribute [rw] full_year
        #   @return [String, nil] the full text description of the academic year
        attr_accessor :full_year

        # @!attribute [rw] lookup
        #   @return [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        attr_accessor :lookup

        # Returns a Year instance for the current academic year
        # @param api [LUSI::API::Core::API] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @return [LUSI::API::Calendar::Year] the current academic year
        # @yield [obj] Passes the Year instance to the block
        # @yieldparam obj [LUSI::API::Calendar::Year] the Year instance
        def self.get_current_academic_year(api = nil, lookup = nil)
          xml = api.call('LUSIReference', 'General.asmx', 'GetCurrentAcademicYear')
          obj = new(LUSI::API::Core::XML.xml_at(xml, 'xmlns:Year'), lookup)
          yield(obj) if block_given?
          obj
        end

        # Returns selected or all defined year instances
        # @param api [LUSI::API::Core::API] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # Remaining positional parameters are years or year identities of the required years
        # @param use_lookup [Boolean] if true, use the lookup service to identify instances before trying the API
        # @return [Array<LUSI::API::Calendar::Year>, nil] the defined year instances
        # @yield [obj] Passes the Year instance to the block
        # @yieldparam obj [LUSI::API::Calendar::Year] the Year instance
        def self.get_instance(api = nil, lookup = nil, *years, use_lookup: true, **kwargs)
          results = nil
          # Search the lookup service if required
          if lookup && use_lookup
            year_lookup = lookup.service(:year)
            results = year_lookup.values if year_lookup
          end
          # Search the API if required
          results = super(api, lookup, 'LUSIReference', 'General.asmx', 'GetYears', 'xmlns:Year') if results.nil?
          # Return the results
          if results.nil? || years.nil? || years.empty?
            # No results or no filtering required
            results
          else
            # Return the year instances for the specified years
            years = years.map { |year| lusi_year_identity(year) }
            results.select { |result| years.include?(result.identity) }
          end
        end

        # Initialises a new Year instance
        # @param (see LUSI::API::Core::Code#initialize)
        # @param full_year [String, nil] the full text description of the academic year
        # @return [void]
        def initialize(xml = nil, lookup = nil, full_year: nil, **kwargs)
          super(xml, lookup, **kwargs)
          self.full_year = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:FullYear', full_year)
          # Store the lookup service for year arithmetic operations
          self.lookup = lookup
        end

        # Returns the year instance n years after this one
        # @param years [Integer] the number of years to add to the instance's year
        # @return [LUSI::API::Calendar::Year] the year instance
        def +(years = 0)
          self.offset(years)
        end

        # Returns the year instance n years before this one
        # @param years [Integer] the number of years to subtract from the instance's year
        # @return [LUSI::API::Calendar::Year] the year instance
        def -(years = 0)
          self.offset(-years)
        end

        # Returns the year instance immediately following this one
        # @return [LUSI::API::Calendar::Year] the year instance
        def next(count = 1)
          count = count.to_i
          if count == 1
            self.offset(1)
          else
            result = []
            (1..count).each { |offset| result.push(self.offset(offset)) }
            result
          end
        end

        # Returns the year instance n years after/before this one
        # @param years [Integer] the number of years to add to the instance's year
        # @return [LUSI::API::Calendar::Year] the year instance
        def offset(years = 0)
          return self if years == 0
          year = self.class.lusi_year(self, self.lookup, as_instance: true, offset: years)
        end

        # Returns the year instance immediately preceding this one
        # @return [LUSI::API::Calendar::Year] the year instance
        def previous(count = 1)
          count = count.to_i
          if count == 1
            self.offset(-1)
          else
            result = []
            (1..count).each { |offset| result.push(self.offset(-offset)) }
            result
          end
        end

        # Returns a list of year instances around (immediately preceding and following) this one
        # @param from [Integer, Range] the starting offset relative to the current year
        # @param to [Integer] the ending offset relative to the current year
        # @return [Array<LUSI::API::Calendar::Year>] the year instances
        def range(from = nil, to = nil)
          from = (from..to) unless from.is_a?(Range)
          result = []
          from.each { |offset| result.push(self.offset(offset)) }
          result
        end

        # Returns a string representation of the Year instance
        # @return [String] the string representation of the Year instance
        def to_s
          self.full_year
        end

        protected

        # Returns an empty parameter hash for the LUSI API call
        def self.get_instance_params(**kwargs)
          {}
        end

      end


    end
  end
end
