require 'lusi_api/core/xml'


module LUSI
  module API
    module Core


      # Represents a LUSI code definition (identity code and text description)
      class BasicCode

        # @!attribute [rw] description
        #   @return [any, nil] the text description
        attr_accessor :description

        # @!attribute [rw] identity
        #   @return [any, nil] the identity code
        attr_accessor :identity

        # Initialises a new Code instance
        # @param xml [Nokogiri::XML::Document, Nokogiri::XML::Node] the XML root of the code from the LUSI API
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param identity [any, nil] the default identity code
        # @param description [String, nil] the default text description
        # @return [void]
        def initialize(xml = nil, lookup = nil, identity: nil, description: nil)
          @description = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Description', description)
          @identity = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Identity', identity)
        end

        # Returns a string representation of the code
        # @return [String, nil] the string representation (text description)
        def to_s
          @description
        end

        # Default get_instance implementation
        # - Instances cannot be retrieved through the LUSI API
        def self.get_instance(*args, **kwargs)
          raise NotImplementedError
        end

        # Default get_instance_params implementation
        # - Instances cannot be retrieved through the LUSI API
        protected

        def self.get_instance_params(*args, **kwargs)
          {}
        end

      end


      # Represents a LUSI code definition which can be retrieved through the LUSI API
      class Code < BasicCode

        # Returns a Code instance for the specifiec LUSI API call
        # @param api [LUSI:API::Core::API] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param path [String] the LUSI API URL path
        # @param endpoint [String] the LUSI API URL endpoint
        # @param method [String] the LUSI API method
        # @param xml_root [String] the XPath of the root element in the LUSI API response
        # @param identity [any] the identity to search for
        # @param description [String, nil] the description to search for
        # @return [Array<LUSI::API::Core::Code>] the matching Code instances
        # @yield [obj] Passes the Code instance to the block
        # @yieldparam obj [LUSI::API::Core::Code] the Code instance
        def self.get_instance(api = nil, lookup = nil, path = nil, endpoint = nil, method = nil, xml_root = nil, **kwargs)
          params = get_instance_params(**kwargs)
          xml = api.call(path, endpoint, method, **params)
          LUSI::API::Core::XML.xml(xml, xml_root) do |c|
            obj = new(c, lookup)
            begin
              yield(obj) if block_given?
            rescue StandardError => e
              puts e
            end
            obj
          end
        end

        protected

        # Returns a hash of parameters for the LUSI API call
        # @param identity [any, nil] the identity code
        # @param description [String. nil] the description
        # @return [Hash<String, any>] the parameter hash for the LUSI API call
        def self.get_instance_params(**kwargs)
          {
            Identity: kwargs[:identity] || '',
            Description: kwargs[:description] || ''
          }
        end

      end


    end
  end
end