require 'lusi_api/core/code'
require 'lusi_api/core/xml'


module LUSI
  module API
    module Country

      # Represents a country in the LUSI API
      class Country < LUSI::API::Core::Code

        # @!attribute [rw] iso_alpha2_code
        #   @return [String, nil] the ISO alpha-2 code for the country
        attr_accessor :iso_alpha2_code

        # Initialises a new Country instance
        # @param (see LUSI::API::Core::Code#initialize)
        # @param iso_alpha2_code [String, nil] the default ISO alpha-2 code for the country
        # @return [void]
        def initialize(xml = nil, lookup = nil, iso_alpha2_code: nil, **kwargs)
          super(xml, lookup, **kwargs)
          @iso_alpha2_code = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:ISOAlpha2Code', iso_alpha2_code)
        end

        # Returns a string representation of the Country instance
        # @return [String] the string representation of the Country instance
        def to_s
          @description
        end

      end


      # Represents the country component of an address in the LUSI API
      class AddressCountry < Country

        # @!attribute [rw] region
        #   @return [String. nil] the text description of the region of the address
        attr_accessor :region

        # Returns AddressCountry instances matching the supplied parameters
        # @param (see LUSI::API::Country::Country#get_instance)
        # @param iso_alpha2_code [String, nil] the ISO alpha-2 code to search for
        # @param region [String, nil] the region to search for
        # @return [Array<AddressCountry>, nil] the matching AddressCountry instances
        # @yield [obj] Passes the AddressCountry instance to the block
        # @yieldparam obj [LUSI::API::Country::AddressCountry] the AddressCountry instance
        def self.get_instance(api = nil, lookup = nil, iso_alpha2_code: nil, region: nil, **kwargs)
          super(api, lookup, 'LUSIReference', 'Lookup.asmx', 'GetAddressCountries', 'xmlns:AddressCountry',
                iso_alpha2_code: iso_alpha2_code, region: region, **kwargs)
        end

        # Initialises a new AddressCountry instance
        # @param (see Country#initialize)
        # @param region [String, nil] the default region
        # @return [void]
        def initialize(xml = nil, lookup = nil, region: nil, **kwargs)
          super(xml, lookup, **kwargs)
          @region = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Region', region)
        end

        protected

        # Returns a parameter hash for the LUSI API call
        # @params (see #get_instance)
        # @return [Hash<String, any] the parameter hash for the LUSI API call
        def self.get_instance_params(**kwargs)
          params = super(**kwargs)
          params[:ISOAlpha2Code] = kwargs[:iso_alpha2_code] || ''
          params[:Region] = kwargs[:region] || ''
          params
        end

      end


      # Represents a nationality in the LUSI API
      # @see Country
      class Nationality < Country

        # Returns Nationality instances matching the specified parameters
        # @param api [LUSI::API::Core::API] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param identity [String, nil] the identity to search for
        # @param description [String, nil] the description to search for
        # @param iso_alpha2_code [String, nil] the ISO alpha-2 code to search for
        # @return [Array<Nationality>, nil] the matching Nationality instances
        # @yield [obj] Passes the Nationality instance to the block
        # @yieldparam obj [LUSI::API::Country::Nationality] the Nationality instance
        def self.get_instance(api = nil, lookup = nil, iso_alpha2_code: nil, **kwargs)
          super(api, lookup, 'LUSIReference', 'Lookup.asmx', 'GetNationalities', 'xmlns:Nationality',
                iso_alpha2_code: iso_alpha2_code, **kwargs)
        end

        protected

        # Returns a parameter hash for the LUSI API call
        # @param (see #get_instance)
        # @param iso_alpha2_code [String, nil] the ISO alpha-2 code to search for
        # @return [Hash<String, any>] the parameter hash for the LUSI API call
        def self.get_instance_params(iso_alpha2_code: nil, **kwargs)
          params = super(**kwargs)
          params[:ISOAlpha2Code] = kwargs[:iso_alpha2_code] || ''
          params
        end

      end

    end
  end
end