require 'lusi_api/core/api'
require 'lusi_api/core/code'
require 'lusi_api/core/xml'


module LUSI
  module API
    module Enrolment

      # Represents an enrolment role in the LUSI API
      class EnrolmentRole < LUSI::API::Core::Code

        # @!attribute [rw] vle_role_description
        #   @return [String, nil] the VLE role description a member has against this enrolment
        attr_accessor :vle_role_description

        # Initialises a new EnrolmentRole instance
        # @param (see LUSI::API::Core::Code#initialize)
        # @param vle_role_description [String, nil] the default VLE role description
        # @return [void]
        def initialize(xml = nil, lookup = nil, vle_role_description: nil, **kwargs)
          super(xml, lookup, **kwargs)
          @vle_role_description = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:VLERoleDescription',
                                                                      vle_role_description)
        end

      end


      # The abstract base class for enrolment classes
      # @abstract Subclasses must define the lusi_ws_* endpoint methods and additional attributes
      class EnrolmentBase

        extend LUSI::API::Core::Endpoint

        # @!attribute [rw] enrolment_role
        #   @return [LUSI::API::Enrolment::EnrolmentRole, nil] the role that members have against this enrolment
        attr_accessor :enrolment_role

        # @!attribute [rw] identity
        #   @return [LUSI::API::Enrolment::EnrolmentIdentity] the identity of the subject (person) of the enrolment
        attr_accessor :identity

        # @!attribute [rw] is_current_enrolment
        #   @return [Boolean, nil] true if the enrolment is current, false otherwise
        attr_accessor :is_current_enrolment

        # @!attribute [rw] is_current_identity
        #   @return [Boolean, nil] true if the member's identity is current, false otherwise
        attr_accessor :is_current_identity

        # @!attribute [rw] username
        #   @return [String, nil] the username of the subject (person) of the enrolment
        attr_accessor :username

        # Initalises a new Enrolment instance
        # @param xml [Nokogiri::XML::Document, Nokogiri::XML::Node] the parsed XML root of the enrolment
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param enrolment_role [LUSI::API::Enrolment::EnrolmentRole, nil] the default enrolment role
        # @param identity [String, nil] the default user identity code
        # @param is_current_enrolment [Boolean, nil] the default current enrolment flag
        # @param is_current_identity [Boolean, nil] the default current identity flag
        # @param username [String, nil] the default username
        # @return [void]
        def initialize(xml = nil, lookup = nil, enrolment_role: nil, identity: nil, is_current_enrolment: nil,
                       is_current_identity: nil, username: nil)
          is_current_enrolment = is_current_enrolment.nil? || is_current_enrolment ? true : false
          is_current_identity = is_current_identity.nil? || is_current_identity ? true : false
          @enrolment_role = EnrolmentRole.new(LUSI::API::Core::XML.xml_at(xml, 'xmlns:EnrolmentRole', enrolment_role),
                                              lookup)
          @identity = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Identity', identity)
          @is_current_enrolment = LUSI::API::Core::XML.xml_boolean_at(xml, 'xmlns:IsCurrentEnrolment',
                                                                      is_current_enrolment)
          @is_current_identity = LUSI::API::Core::XML.xml_boolean_at(xml, 'xmlns:IsCurrentIdentity',
                                                                     is_current_identity)
          @username = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Username', username)
        end

        # Returns an array of instances matching the specified search criteria
        # @param api [LUSI::API::Core::API] the LUSI API instance to use for searching
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param (see #get_instance_params)
        # @yield [obj] Passes the instance to the block
        def self.get_instance(api, lookup = nil, current_only: nil, **kwargs)
          current_only = current_only.nil? || current_only ? true : false
          if current_only
            # Filter nodes which have current enrolments and user identities
            filter = Proc.new do |node|
              LUSI::API::Core::XML.xml_boolean_at(node, 'xmlns:IsCurrentEnrolment', false) &&
                  LUSI::API::Core::XML.xml_boolean_at(node, 'xmlns:IsCurrentIdentity', false)
            end
          else
            filter = nil
          end
          params = self.get_instance_params(**kwargs)
          xml = api.call(self.lusi_ws_path, self.lusi_ws_endpoint, self.lusi_ws_method, **params)
          result = LUSI::API::Core::XML.xml(xml, "xmlns:#{self.lusi_ws_xml_root}", filter: filter) do |m|
            obj = self.new(m, lookup)
            yield(obj) if block_given?
            obj
          end
          # Return the array of instances
          result
        end

        # Returns the lookup indices supported by this enrolment type
        # @return [Array<Symbol>] the supported lookup indices
        def lookup_indices
          [:identity, :role, :username]
        end

        # Returns the value to be used as a key for the specified lookup index
        # @return [Object] the key value
        def lookup_key(index = nil)
          case index
            when :identity
              self.identity
            when :role
              self.enrolment_role ? self.enrolment_role.identity : nil
            when :username
              self.username
            else
              nil
          end
        end

        # @see (LUSI::API::Core::Endpoint#lusi_ws_path)
        def self.lusi_ws_path
          'UserDetails'
        end

        protected

        # Returns a hash of parameters for the LUSI API call. Subclasses may extend or override this method.
        # @param active_vle_space_only [Boolean, nil] if true, restrict results to active VLE spaces only
        # @param asp_identity [String, nil] return instances matching the academically significant period (ASP) identity
        # @param cohort_identity [String, nil] return instances matching the cohort identity
        # @param course_identity [String, nil] return instances matching the course identity
        # @param current_student_only [Boolean, nil] if true, restrict results to current students only
        # @param department_identity [String, nil] return instances matching the department identity
        # @param require_username [Boolean, nil] if true, include the student username, ignore students with no username
        # @param year_identity (String, nil) return instances matching the year identity
        # @return [Hash<String, String>] the parameters for the LUSI API call
        def self.get_instance_params(**kwargs)
          result = {
            ActiveVLESpaceOnly: kwargs.fetch(:active_vle_space_only, true) ? 'true' : 'false',
            ASPIdentity: kwargs.fetch(:asp_identity, ''),
            CohortIdentity: kwargs.fetch(:cohort_identity, ''),
            CourseIdentity: kwargs.fetch(:course_identity, ''),
            DepartmentIdentity: kwargs.fetch(:department_identity, ''),
            RequireUsername: kwargs.fetch(:require_username, true) ? 'true' : 'false',
            YearIdentity: kwargs.fetch(:year_identity, '')
          }
          if self.lusi_ws_endpoint == 'Student.asmx'
            result[:CurrentStudentOnly] = kwargs.fetch(:current_student_only, true) ? 'true' : 'false'
          end
          result
        end

      end


      class EnrolmentLookup

        # Initialises a new EnrolmentLookup instance
        def initialize(enrolments = nil, *indices, &block)
          @default_proc = block
          @indices = {}
          enrolments.each { |e| add(e, *indices) }
        end

        # @see (LUSI::API::Enrolment::EnrolmentLookup#fetch)
        def [](key, *indices)
          fetch(key, *indices)
        end

        # Adds an enrolment to specified indices (default is all indices if unspecified)
        # @param enrolment [LUSI::API::Enrolment::EnrolmentBase] the enrolment to add
        # Reminaing positional parameters specify the indices to update
        # @return [void]
        def add(enrolment = nil, *indices)
          indices = enrolment.lookup_indices if indices.nil? || indices.empty?
          indices.each { |index| add_enrolment(enrolment, index) }
          nil
        end

        # Searches the specified indices for the lookup key and returns the first match
        # @param key [Object] the lookup key. If the key defines method #enrolment_lookup_keys, this method determines
        #   which keys are searched for.
        # Remaining positional parameters specify the indices to search. If no indices are specified, but the key
        # instance defines method #enrolment_lookup_indices, this method determines which indices are searched.
        # @return [Array<LUSI::API::Enrolment::EnrolmentBase>, nil] the enrolments corresponding to key, or nil
        #   if match was found
        def fetch(key = nil, *indices)

          # If no index is specified, infer it from the key type if possible
          if indices.nil? || indices.empty?
            indices = key.respond_to?(:enrolment_lookup_indices) ? key.enrolment_lookup_indices : nil
          end
          return nil if indices.nil? || indices.empty?

          # Use the lookup keys specified by the key instance if possible, otherwise use the literal key value
          keys = key.respond_to?(:enrolment_lookup_keys) ? key.enrolment_lookup_keys : [key]

          # Search the specified indices until a match is found, then return matches for all key values from this index
          unless keys.nil? || keys.empty?
            result = []
            indices.each do |index|
              i = @indices[index]
              if i
                keys.each { |key| result += i[key] || [] }
              end
              return result unless result.empty?
            end
          end

          # If we get here, the search failed in all indices - call the default_proc if available, otherwise return nil
          @default_proc ? @default_proc.call(key) : nil

        end

        protected

        # Adds an enrolment to the specified index
        # @param enrolment [LUSI::API::Enrolment::EnrolmentBase] the enrolment
        # @param index [Symbol] the index to be updated
        def add_enrolment(enrolment, index)
          # Get the specified index hash
          @indices[index] = {} unless @indices.include?(index)
          hash = @indices[index]
          # Add the enrolment to the list
          key = enrolment.lookup_key(index)
          if hash.include?(key)
            hash[key].push(enrolment)
          else
            hash[key] = [enrolment]
          end
          # Return the list containing the enrolment
          hash[key]
        end

      end


    end
  end
end