require 'lusi_api/core/code'
require 'lusi_api/core/xml'


module LUSI
  module API
    module Person


      # Represents a staff course role in the LUSI API
      class StaffCourseRole < LUSI::API::Core::Code

        # @!attribute [rw] vle_role_description
        #   @return [String, nil] the VLE role description
        attr_accessor :vle_role_description

        # Returns an instance of StaffCourseRole matching the specified parameters
        # @param api [LUSI::API::Core::API, nil] the LUSI API instance
        # @param lookup [LUSI::API::Core::Lookup::LookupService, nil] the lookup service for object resolution
        # @param (see LUSI::API::Core::Code#get_instance)
        # @return [Array<LUSI::API::Person::StaffCourseRole>, nil]
        def self.get_instance(api = nil, lookup = nil, **kwargs)
          super(api, lookup, 'LUSIReference', 'Lookup.asmx', 'GetStaffCourseRoles', 'xmlns:StaffCourseRole',
                result_class: StaffCourseRole, **kwargs)
        end

        # Initialises a new StaffCourseRole instance
        # @param (see LUSI::API::Core::Code#initialize)
        # @param vle_role_description [String, nil] the default VLE role description
        # @return [void]
        def initialize(xml = nil, lookup = nil, vle_role_description: nil, **kwargs)
          super(xml, lookup, **kwargs)
          @vle_role_description = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:VLERoleDescription',
                                                                      vle_role_description)
        end

      end


      # Represents a staff member in the LUSI API
      class StaffMember

        # @!attribute [rw] identity
        #   @return [String, nil] the identity code of the staff member
        attr_accessor :identity

        # @!attribute [rw] title
        #   @return [String, nil] the title (honourific) of the staff member
        attr_accessor :title

        # @!attribute [rw] surname
        #   @return [String, nil] the surname of the staff member
        attr_accessor :surname

        # @!attribute [rw] forename
        #   @return [String, nil] the forename of the staff member
        attr_accessor :forename

        # @!attribute [rw] preferred_name
        #   @return [String, nil] the preferred name of the staff member
        attr_accessor :preferred_name

        # @!attribute [rw] department
        #   @return [LUSI::API::Organisation::Unit, nil] the staff member's associated department
        attr_accessor :department

        # @!attribute [rw] faculty
        #   @return [LUSI::API::Organisation::Unit, nil] the staff member's associated faculty
        attr_accessor :faculty

        # @!attribute [rw] institution
        #   @return [LUSI::API::Organisation::Unit, nil] the staff member's associated institution
        attr_accessor :institution

        # Initialises a new StaffMember instance
        # @param xml [Nokogiri::XML::Document, Nokogiri::XML::Node] the parsed XML root of the staff member
        # @param lookup [LUSI::API::Core::LookupTable, nil] the lookup service for organisation unit resolution
        # @param department [LUSI::API::Organisation::Unit, nil] the default department
        # @param faculty [LUSI::API::Organisation::Unit, nil] the default faculty
        # @param forename [String, nil] the default forename
        # @param identity [String, nil] the default identity code
        # @param institution [LUSI::API::Organisation::Unit, nil] the default institution
        # @param preferred_name [String, nil] the default preferred name
        # @param surname [String, nil] the default surname
        # @param title [String, nil] the default title (honourific)
        # @return [void]
        def initialize(xml = nil, lookup = nil, department: nil, faculty: nil, forename: nil, identity: nil,
                       institution: nil, preferred_name: nil, surname: nil, title: nil)
          @department = LUSI::API::Core::XML.lookup(xml, lookup, :department, 'xmlns:Department/xmlns:Identity',
                                                    department)
          @faculty = LUSI::API::Core::XML.lookup(xml, lookup, :faculty, 'xmlns:Faculty/xmlns:Identity', faculty)
          @forename = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Forename', forename)
          @identity = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Identity', identity)
          @institution = LUSI::API::Core::XML.lookup(xml, lookup, :institution, 'xmlns:Institution/xmlns:Identity',
                                                     institution)
          @preferred_name = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:PreferredName', preferred_name)
          @surname = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Surname', surname)
          @title = LUSI::API::Core::XML.xml_content_at(xml, 'xmlns:Title', title)
        end

        # Returns a string representation of the staff member (full name)
        # @return [String] the string representation of the staff member (full name)
        def to_s
          str = "#{@title} #{@forename} #{@surname}"
          str.strip!
          str.squeeze!(' ')
          str
        end

      end


    end
  end
end
